package ch.fritteli.labyrinth.generator.renderer.textfile;

import ch.fritteli.labyrinth.generator.model.Labyrinth;
import ch.fritteli.labyrinth.generator.renderer.Renderer;
import ch.fritteli.labyrinth.generator.renderer.text.TextRenderer;
import io.vavr.collection.List;
import lombok.NonNull;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;

public class TextFileRenderer implements Renderer<List<Path>> {
    @NonNull
    private static final TextRenderer TEXT_RENDERER = TextRenderer.newInstance();
    private Path targetLabyrinthFile;
    private Path targetSolutionFile;

    private TextFileRenderer() {
        try {
            this.targetLabyrinthFile = Files.createTempFile("labyrinth_", ".txt");
            this.targetSolutionFile = this.targetLabyrinthFile.getParent().resolve(
                    this.targetLabyrinthFile.getFileName().toString().replace(".txt", "-solution.txt")
            );
        } catch (IOException e) {
            System.err.println("Unable to set default target file.");
            e.printStackTrace();
        }
    }

    @NonNull
    public static TextFileRenderer newInstance() {
        return new TextFileRenderer();
    }

    public boolean isTargetLabyrinthFileDefinedAndWritable() {
        return this.targetLabyrinthFile != null && this.targetLabyrinthFile.toFile().canWrite();
    }

    public boolean isTargetSolutionFileDefinedAndWritable() {
        return this.targetSolutionFile != null && this.targetSolutionFile.toFile().canWrite();
    }

    @NonNull
    public TextFileRenderer setTargetLabyrinthFile(@NonNull final Path targetLabyrinthFile) {
        this.targetLabyrinthFile = targetLabyrinthFile;
        return this;
    }

    @NonNull
    public TextFileRenderer setTargetSolutionFile(@NonNull final Path targetSolutionFile) {
        this.targetSolutionFile = targetSolutionFile;
        return this;
    }

    @Override
    public List<Path> render(@NonNull final Labyrinth labyrinth) {
        if (!this.isTargetLabyrinthFileDefinedAndWritable()) {
            try {
                Files.createFile(this.targetLabyrinthFile);
            } catch (IOException e) {
                throw new IllegalArgumentException("Cannot write to target labyrinth file.", e);
            }
        }
        if (!this.isTargetSolutionFileDefinedAndWritable()) {
            try {
                Files.createFile(this.targetSolutionFile);
            } catch (IOException e) {
                throw new IllegalArgumentException("Cannot write to target solution file.", e);
            }
        }
        final String text;
        final String solution;
        synchronized (TEXT_RENDERER) {
            text = TEXT_RENDERER.setRenderSolution(false).render(labyrinth).strip();
            solution = TEXT_RENDERER.setRenderSolution(true).render(labyrinth).strip();
        }
        try {
            Files.write(this.targetLabyrinthFile, text.getBytes(StandardCharsets.UTF_8));
            Files.write(this.targetSolutionFile, solution.getBytes(StandardCharsets.UTF_8));
        } catch (IOException e) {
            System.err.println("Failed writing to file " + this.targetLabyrinthFile.normalize().toString());
            e.printStackTrace();
        }
        return List.of(this.targetLabyrinthFile, this.targetSolutionFile);
    }
}
